from typing import Any, Callable, Union

Predicate = Callable[[Any], bool]


class QvalValidationError(Exception):
    """
    An error raised on if validation fails.
    This exception should be used to provide custom validation error message to the frontend.

    Example:
        >>> from qval import validate
        >>> def f(v: str) -> bool:
        ...     if not v.isnumeric():
        ...         raise QvalValidationError(f"Expected a number, got \'{v}\'")
        ...     return True
        >>> params = validate({"number": "42"}, {"number": f})
        >>> with params: pass  # OK
        >>> with params.apply_to_request({"number": "a string"}): pass
        Traceback (most recent call last):
            ...
        qval.exceptions.InvalidQueryParamException: ...
    """


class Validator(object):
    """
    Validates given value using provided predicates.

    .. automethod:: __call__
    """

    # :class:`Validator` implements __call__(Any) -> bool,
    # therefore can be treated in the same way as :type:`ValidatorType`.
    # For the sake of clarity, it is reflected in the class attributes below.
    ValidatorType = Union["Validator", Predicate]
    Predicate = ValidatorType

    def __init__(self, *predicates: Predicate):
        """
        Instantiates the validator.

        :param predicates: predefined predicates
        :type predicates: Callable[[Any], bool]
        """
        self.predicates = list(predicates)

    def add(self, predicate: Predicate) -> "Validator":
        """
        Adds new predicate to the list.

        :param predicate: predicate function
        :return: self
        """
        self.predicates.append(predicate)
        return self

    def __call__(self, value: Any) -> bool:
        """
        Applies all stored predicates to the given value.

        :param value: value to validate
        :return: True if all checks have passed, False otherwise
        """
        for p in self.predicates:
            if not p(value):
                return False
        return True



from decimal import Decimal

from django.views.generic import DetailView
from django.http import HttpRequest, JsonResponse
from qval import qval, validate, Validator, QvalValidationError


def division_view(request: HttpRequest):
    """
    GET /api/divide?
    param a : int
    param b : int, nonzero

    Example: GET /api/divide?a=10&b=2&token=abcdefghijkl -> 200, {"answer": 5}
    """
    # Parameter validation occurs in the context manager.
    # If validation fails or user code throws an error, context manager
    # will raise InvalidQueryParamException or APIException respectively.
    # In Django, these exception will be processed and result
    # in error codes (400 and 500) on the client side.
    params = (
        validate(request, a=int, b=int)
        # `b` must be anything but zero
        .nonzero("b")
    )
    with params as p:
        return JsonResponse({"answer": p.a // p.b})


@qval({"a": float, "b": float})
def pow_view(request, params):
    """
    GET /api/pow?
    param a : float
    param b : float

    Example: GET /api/pow?a=2&b=3 -> 200, {"answer": 8.0}
    Example: GET /api/pow?a=2&b=3000000000000 -> 500,
    {
        "error": "An error occurred while processing you request. Please contact the website administrator."
    }
   """
    # Here we don't catch the OverflowError if `b` is too big.
    # This will result in 500 error on the client side.
    return JsonResponse({"answer": params.a ** params.b})


class PurchaseView(DetailView):
    @staticmethod
    def price_validator(price: int) -> bool:
        """
        A predicate to validate `price` query parameter.
        Provides custom error message.
        """
        if price <= 0:
            # If price does not match our requirements, we raise QvalValidationError() with a custom message.
            # This exception will be handled in the context manager and will be reraised
            # as InvalidQueryParamException() [HTTP 400].
            raise QvalValidationError(
                f"Price must be greater than zero, got '{price}'."
            )
        return True

    purchase_factories = {"price": Decimal, "item_id": int, "token": None}
    purchase_validators = {
        "token": Validator(lambda x: len(x) == 12),
        # Validator(p) can be omitted if there is only one predicate:
        "item_id": lambda x: x >= 0,
        # Access underlying function without get/set protocol.
        # In order to avoid this hack, define validators outside of the class.
        "price": price_validator.__func__,
    }

    @qval(purchase_factories, purchase_validators)
    def get(self, request, params):
        """
        GET /api/purchase?
        param item_id : int, positive
        param price   : float, greater than zero
        param token   : string, length == 12

        Example: GET /api/purchase?item_id=1&price=5.8&token=abcdefghijkl
                 -> {"success": "Item '1' has been purchased. Check: 5.92$."
        """
        tax = 0.02
        cost = params.price * Decimal(1 + tax)
        return JsonResponse(
            {
                "success": f"Item '{params.item_id}' has been purchased. Check: {round(cost, 2)}$."
            }
        )


__all__ = ["division_view", "pow_view", "PurchaseView"]



import pytest

from qval import InvalidQueryParamException
from qval.utils import make_request
from qval.qval import QueryParamValidator
from tests.crossframework import builder


def test_chain_validators():
    request = {
        "num": "42",
        "num2": "-10",
        "num3": "3.14",
        "power": "16",
        "power2": "25",
        "new_param": "2.79",
    }
    # fmt: off
    factories = {
        "num": int,
        "num2": int,
        "num3": float,
        "power": int,
        "power2": lambda x: float(x) ** 0.5
    }
    # fmt: on

    qval = (
        QueryParamValidator(make_request(request), factories)
        .check("power", lambda x: (x ** 0.5).is_integer())
        .check("power2", float.is_integer)
        .eq("num", 42)
        .gt("num", 41)
        .lt("num2", 0)
        .nonzero("num3")
        .positive("power")
    )
    qval.add_predicate(
        "new_param", lambda x: float(x) not in (float("inf"), float("-inf"))
    )
    for r in builder.iterbuild(request):
        params = qval.apply_to_request(r)
        with params as p:
            assert p.num == 42
            assert p.num2 == -10
            assert p.num3 == 3.14
            assert p.power == 16
            assert p.power2 == 5
            assert p.new_param == "2.79"

    bad_examples = {
        "num": "0",  # is not equal to 42
        "num2": "1",  # is greater than zero
        "num3": "0",  # is equal to zero
        "power": "-64",  # is not positive
        "power2": "24",  # is not a perfect square
        "new_param": "inf",  # is an infinity
    }

    for r in builder.iterbuild(bad_examples):
        params = qval.apply_to_request(r)
        with pytest.raises(InvalidQueryParamException) as e, params:
            pass
        assert e.type is InvalidQueryParamException



import os
from importlib import reload, import_module

import pytest


@pytest.fixture()
def isolation():
    os.environ["DJANGO_SETTINGS_MODULE"] = "tests.test_make_request_wrapper"
    yield None
    del os.environ["DJANGO_SETTINGS_MODULE"]


def test_make_request_wrapper(isolation):
    _ = reload(import_module("qval.framework_integration"))
    make_request = reload(import_module("qval.utils")).make_request
    r = {"param": "value"}
    _ = make_request(r)
    assert make_request.__list__[0] == r


def make_request_wrapper(f):
    def wrapper(request):
        wrapper.__list__.append(request)
        return f(request)

    wrapper.__list__ = []
    return wrapper


QVAL_MAKE_REQUEST_WRAPPER = make_request_wrapper
